#include "osd.h"
#include "common.h"
#include <string.h>

typedef unsigned char *PBYTE;
#include <fmp.h>
#include <rm84mpegcmn.h>
#include <hwiface.h>

//#define EM84_HAS_4MB  1
extern bool mOSDHas4MB;

#define COL_A 0xFF000000
#define COL_R 0x00FF0000
#define COL_G 0x0000FF00
#define COL_B 0x000000FF

#define COL_A_SH 24
#define COL_R_SH 16
#define COL_G_SH  8
#define COL_B_SH  0


#define EM84OSD_MAX_WIDTH  720
#define EM84OSD_MAX_HEIGHT 576
#define EM84OSD_MAX_COLORS 256

#define EM84OSD_HEADERSIZE       8
#define EM84OSD_SIZE_PALETTE     (EM84OSD_MAX_COLORS * 4)
#define EM84OSD_OFFSET_BMP       (EM84OSD_HEADERSIZE + EM84OSD_SIZE_PALETTE)

#define EM84OSD_HWBUFSIZE (mOSDHas4MB ? 2000000 : 41000)

#define EM84OSD_MAX_BUFFERSIZE (EM84OSD_OFFSET_BMP + (EM84OSD_MAX_WIDTH * EM84OSD_MAX_HEIGHT))

class cOsdEm84xx: public cOsd {
private:

   IOsd *mOsd;
   unsigned char *mOsdBuffer;
   unsigned char *mCOsdBuffer;
   tColor *mPaletteEM84;
   int mUsedColors;
   int mCurwidth;
   int mCurheight;
   bool mUseHilite;
   int mCompLines;
   int mHilitePal[2][4]; // color Palette for nonhilite and hilite Area
   int mMapColor[2][8];

   int mHiliteNum[2];    // number of nonhilite and hilite Areas
   OSD_WND mHiliteArea[2]; // coordinates of nonhilite and hilite Area ( x,y,width,height)
   int mHiliteIdx;    // Index of Hilite area
   int mNoHiliteIdx;    // Index of NonHilite area
   int minX;
   int minY;
   int maxX;
   int maxY;

protected:
   void displayOSD(int x, int y);
   bool CheckHiliteArea( int *newColors, int numColors, cBitmap *Bitmap );
   void commitPaletteToOsd();
   int suiteColorToPalette(tColor vdrcolor);
   static void rgbtoyuv(double R,double G,double B,double *y,double *u,double *v);
   int Compress2(int size);
   int Compress4(int size);
   unsigned char GetHiliteColor( unsigned char idx, int x, int y );
   
public:
   cOsdEm84xx(int Left, int Top);
   virtual ~cOsdEm84xx();

   virtual eOsdError CanHandleAreas(const tArea *Areas, int NumAreas);
   //virtual eOsdError SetAreas(const tArea *Areas, int NumAreas);
   virtual void Flush(void);
};



cOsdEm84xx::cOsdEm84xx(int Left, int Top):
cOsd(Left, Top)
{
   Dprintf("%s\n", __PRETTY_FUNCTION__);

   mOsd = NULL;
   mOsdBuffer = NULL;
   mCOsdBuffer = NULL;
   mPaletteEM84 = NULL;
   mCurwidth = EM84OSD_MAX_WIDTH - Left;
   mCurheight = EM84OSD_MAX_WIDTH - Top;

   if (FMPQueryInterface(IID_IOSD, (void**)&mOsd)) {
      Dprintf("could not open OSD\n");
      mOsd = NULL;
   }
}

cOsdEm84xx::~cOsdEm84xx()
{
   if (mOsd)
      mOsd->lpVtbl->Off(mOsd);
      
   if(mPaletteEM84 != NULL)
      free(mPaletteEM84);
   if(mOsdBuffer != NULL)
      free(mOsdBuffer);
   if(mCOsdBuffer != NULL)
      free(mCOsdBuffer);
      
   //mOsd->lpVtbl->Flush(mOsd);  
}


eOsdError cOsdEm84xx::CanHandleAreas(const tArea *Areas, int NumAreas)
{
   Dprintf("%s\n", __PRETTY_FUNCTION__);
   eOsdError Result = cOsd::CanHandleAreas(Areas, NumAreas);
   if (Result == oeOk) {
      for (int i = 0; i < NumAreas; i++) {
         if (Areas[i].bpp != 1 && Areas[i].bpp != 2 && Areas[i].bpp != 4 && Areas[i].bpp != 8)
            return oeBppNotSupported;
         if (Areas[i].Width() < 1 || Areas[i].Height() < 1 || Areas[i].x2 > EM84OSD_MAX_WIDTH || Areas[i].y2 > EM84OSD_MAX_HEIGHT)
            return oeWrongAreaSize;
         Dprintf("Area %d %d-%d-%d-%d\n", i, Areas[i].x1,Areas[i].Width(),Areas[i].y1,Areas[i].Height() );
      }
   }
   return Result;
}


void cOsdEm84xx::Flush(void)
{
   Dprintf("%s\n", __PRETTY_FUNCTION__);

   cBitmap *Bitmap;

   if (!mOsd)
      return;

   if(mOsdBuffer == NULL)
      mOsdBuffer = (unsigned char *)malloc(EM84OSD_MAX_BUFFERSIZE); 

   if(mPaletteEM84 == NULL)
      mPaletteEM84 = (tColor *)malloc(sizeof(tColor)*EM84OSD_MAX_COLORS); 

   if (!mOSDHas4MB)
      mUseHilite = true;
   else   
      mUseHilite = false;
      
   mCompLines = 0;
   memset(mOsdBuffer, 0, EM84OSD_MAX_BUFFERSIZE);
   memset(mPaletteEM84, 0, (sizeof(tColor)*EM84OSD_MAX_COLORS));
   
   mOsd->lpVtbl->Flush(mOsd);

   mUsedColors = 0;

   if (mUseHilite) {
      for( int i = 0; i < 4; i++ ) { 
         mHilitePal[0][i] = mHilitePal[1][i] = -1;
      }
      for( int i = 0; i < 8; i++ ) { 
         mMapColor[0][i] = mMapColor[1][i] = 0;
      }
      mHiliteArea[0].x = mHiliteArea[1].x = 0;
      mHiliteArea[0].y = mHiliteArea[1].y = 0;
      mHiliteArea[0].w = mHiliteArea[1].w = 0;
      mHiliteArea[0].h = mHiliteArea[1].h = 0;
   
      mHiliteNum[0] = mHiliteNum[1] = 0;
      mNoHiliteIdx = 0;
      mHiliteIdx = 1;
   }
   
   minX = EM84OSD_MAX_WIDTH + 1;
   minY = EM84OSD_MAX_HEIGHT + 1;

   maxX = maxY = 0;

   // first build a common palette (don't know if this will work !?!?
   for (int i = 0; (Bitmap = GetBitmap(i)) != NULL; i++) {

      int x1 = 0, y1 = 0, x2 = 0, y2 = 0;
      if (Bitmap->Dirty(x1, y1, x2, y2)) {
         // commit colors:
 
         Dprintf("Area %d %d-%d-%d-%d\n", i, x1,x2,y1,y2 );
         int NumColors;
         const tColor *Colors = Bitmap->Colors(NumColors);
         if (Colors) {
            int* newColors = (int*)malloc(NumColors*sizeof(int));
            
            for (int j = 0; j < NumColors; j++) {
               newColors[j] = suiteColorToPalette(Colors[j]);
            }
            
            if (mUseHilite)
               mUseHilite = CheckHiliteArea( newColors, NumColors, Bitmap );
            
            // draw window to osd!
            int x,y, cx, cy;
            for (y = y1; y <= y2; y++) {
               for (x = x1; x <= x2; x++) {
                  cx = Bitmap->X0()+x;
                  cy = Bitmap->Y0()+y;
                  if (cx < mCurwidth && cy < mCurheight)
                     mOsdBuffer[EM84OSD_OFFSET_BMP+(cy*mCurwidth+cx)] = newColors[*((unsigned char*)(Bitmap->Data(x,y)))];
               }
            }
            if((Bitmap->X0()+x1) < minX)
               minX = Bitmap->X0()+x1;
            if((Bitmap->X0()+x2) > maxX)
               maxX = Bitmap->X0()+x2;
         
            if((Bitmap->Y0()+y1) < minY)
               minY = Bitmap->Y0()+y1;
            if((Bitmap->Y0()+y2) > maxY)
               maxY = Bitmap->Y0()+y2;
         }
      }
   }
   Dprintf("%d-%d-%d-%d\n", minX, minY, maxX, maxY );
//   if( maxY < mCurheight )
//      mCurheight = maxY;

   Dprintf("MCW %d-%d\n", mCurwidth, mCurheight );
      
   commitPaletteToOsd();
   displayOSD(Left(), Top());
}

int cOsdEm84xx::suiteColorToPalette(tColor vdrcolor)
{
   int difr, difg, difb;
   int alldif;
   int i;
   int start = 0;
   int stop = mUsedColors;

   for (i=start ; i < stop; i++) {
      if ( mPaletteEM84[i] == vdrcolor ){
         return i;
      }
   }

   if (mUsedColors < EM84OSD_MAX_COLORS) {
      Dprintf("Color: %d - %X\n", mUsedColors, vdrcolor );
      mPaletteEM84[mUsedColors] = vdrcolor;
      mUsedColors++;
      return(mUsedColors-1);
   }

   // choose best one! ;)
   int found = -1;
   alldif = 3*256;

   start = 0;
   stop  = EM84OSD_MAX_COLORS;

   for (i=start; i < stop; i++) {
      difr = (( mPaletteEM84[i] & vdrcolor ) & COL_R ) >> COL_R_SH;
      difg = (( mPaletteEM84[i] & vdrcolor ) & COL_G ) >> COL_G_SH;
      difb = (( mPaletteEM84[i] & vdrcolor ) & COL_B ) >> COL_B_SH;
      if (difr < 0)
         difr = -1*difr;
      if (difg < 0)
         difg = -1*difg;
      if (difb < 0)
         difb = -1*difb;

      if ((difr + difg + difb) < alldif) {
         alldif = (difr + difg + difb);
         found = i;
      }
   }
   Dprintf("Color: %X mapped: %d\n", vdrcolor, found );

   return found;
}


unsigned char cOsdEm84xx::GetHiliteColor( unsigned char idx, int x, int y )
{
   int colIdx;
   if( x >= mHiliteArea[mHiliteIdx].x && x <= ( mHiliteArea[mHiliteIdx].x + mHiliteArea[mHiliteIdx].w ) &&
       y >= mHiliteArea[mHiliteIdx].y && y <= ( mHiliteArea[mHiliteIdx].y + mHiliteArea[mHiliteIdx].h ) )
       colIdx = mHiliteIdx;
   else 
       colIdx = mNoHiliteIdx;

   return mMapColor[colIdx][idx];
}

void cOsdEm84xx::commitPaletteToOsd()
{
   int i;
   double y,u,v;
   tColor color;
   unsigned char* ptr = &mOsdBuffer[EM84OSD_HEADERSIZE];

   if( mUseHilite ) {
      if( mUsedColors < 5 )
         mUseHilite = false;
      else
         mUsedColors = 8;   
   }
            
   for (i=0; i < mUsedColors; i++) {
      if( mUseHilite ) {
         if( i < 4 ) {
            mMapColor[mNoHiliteIdx][mHilitePal[mNoHiliteIdx][i]] = i;
            color = mPaletteEM84[mHilitePal[mNoHiliteIdx][i]];
         }
         else {
            mMapColor[mHiliteIdx][mHilitePal[mHiliteIdx][i]] = i;
            color = mPaletteEM84[mHilitePal[mHiliteIdx][i]];
         }   
      }     
      else
         color = mPaletteEM84[i];
         
      rgbtoyuv((double)((color & COL_R ) >> COL_R_SH),
               (double)((color & COL_G ) >> COL_G_SH),
               (double)((color & COL_B ) >> COL_B_SH),
               &y,&u,&v);

      *ptr++ = (unsigned char)(( color & COL_A ) >> COL_A_SH);
      *ptr++ = (unsigned char) y;
      *ptr++ = (unsigned char) u;
      *ptr++ = (unsigned char) v;
   }
   
   if( !mUseHilite ) {
      if( mUsedColors < 5 )
         mUsedColors = 4;
      else if( mUsedColors < 17 )
         mUsedColors = 16;
      else
         mUsedColors = 256; 
   }
}

void cOsdEm84xx::displayOSD(int x, int y)
{
   Dprintf("%s(%d,%d)\n", __PRETTY_FUNCTION__, x, y);

   if (mOsd ) {
      int size = (mCurwidth * mCurheight) + EM84OSD_OFFSET_BMP;
      if ( size > EM84OSD_HWBUFSIZE ) {
         int cOffset = (mUsedColors * 4) + EM84OSD_HEADERSIZE;
         
         if(mCOsdBuffer == NULL)
            mCOsdBuffer = (unsigned char *)malloc(EM84OSD_HWBUFSIZE); 

         memset(mCOsdBuffer,0x00,EM84OSD_HWBUFSIZE);
         
         if ( mUsedColors < 9 ) {
            size = Compress2(EM84OSD_HWBUFSIZE - cOffset);
            if ( mUsedColors < 5 )
               mOsdBuffer[0] = 0x09;
            else   
               mOsdBuffer[0] = 0x11;
         }
         else {
            if ( mUsedColors > 16 ) {
               // do anything for reducing colors ...
               mUsedColors = 16;
            }
            size = Compress4(EM84OSD_HWBUFSIZE - cOffset);
            mOsdBuffer[0] = 0x13;
         }
         memcpy(mOsdBuffer + cOffset, mCOsdBuffer, size);

         if ( mCompLines > 0 && mCompLines < mCurheight ) {
            Dprintf("Comp: (%d,%d)\n", mCompLines, mCurheight);
            mCurheight = mCompLines;
         }
         if ( size > EM84OSD_HWBUFSIZE ) {
            // for further use: enable splashosd !!
         }
      }
      else
         mOsdBuffer[0] = 0x3e;
            
      
      mOsdBuffer[1] = size >> 16;
      mOsdBuffer[2] = (size >> 8) & 0xff;
      mOsdBuffer[3] = size & 0xff;
      mOsdBuffer[4] = mCurwidth >> 8;
      mOsdBuffer[5] = mCurwidth & 0xff;
      mOsdBuffer[6] = mCurheight >> 8;
      mOsdBuffer[7] = mCurheight & 0xff;
      OSD_WND *WndHilite;
      if (mUseHilite) {
         WndHilite = &(mHiliteArea[mHiliteIdx]);
      }
      else {
         WndHilite = &(mHiliteArea[0]);
         WndHilite->x = 0;
         WndHilite->y = 0;
         WndHilite->w = 0;
         WndHilite->h = 0;
      }
      mOsd->lpVtbl->SetHli(mOsd, WndHilite);
      Dprintf("HiLite(%d,%d,%d,%d)\n", WndHilite->x, WndHilite->y, WndHilite->w, WndHilite->h);

      OSD_WND Wnd;
      Wnd.x = x;
      Wnd.y = y;
      Wnd.w = mCurwidth;
      Wnd.h = mCurheight;
      Dprintf("Win(%d,%d,%d,%d) - %d\n", Wnd.x, Wnd.y, Wnd.w, Wnd.h, size);

      mOsd->lpVtbl->Write(mOsd , mOsdBuffer, size);
      mOsd->lpVtbl->SetDest(mOsd , &Wnd);

      //OSSleep(10);
      mOsd->lpVtbl->On(mOsd );
   }
}


int cOsdEm84xx::Compress2( int oBufSize )
{
   int iBufSize = (mCurwidth * mCurheight) + EM84OSD_OFFSET_BMP;
   unsigned char b,b1,b2, oldb, newb, incomplete, temp;
   int iIdx, oIdx, count, newline, size;
   unsigned char *iBuf = mOsdBuffer;
   unsigned char *oBuf = mCOsdBuffer;

   oIdx = 0;
   mCompLines = 0;
   iIdx = EM84OSD_OFFSET_BMP;
   size = 0;
   newline = 0;
   incomplete = 0;
   if( mUsedColors > 4 )
      b = GetHiliteColor( iBuf[0], mCompLines, size );
   else
      b = iBuf[0];
   
   while (iIdx < iBufSize && oIdx < oBufSize) {
      oldb = b;
      count = 0;
      while ((iIdx < iBufSize) && (count < 60) && (b == oldb) && (size < mCurwidth)) {
         count++;
         oldb = b;
         if( mUsedColors > 4 )
            b = GetHiliteColor( iBuf[0], mCompLines, size );
         else
            b = iBuf[0];
            
         iIdx++;

         if (newline) {
            oldb = b;
            newline = 0;
         }

         size++;
      }

      if (size == mCurwidth) {
         b1 = oldb & 0x03;
         newb = 0;
         newb = newb | b1;
         size = 0;
         if (incomplete == 1) {
            temp = (newb >> 4) & 0x0F;
            oBuf[oIdx] = oBuf[oIdx] | temp;
            oIdx++;
            temp = newb & 0x0F;
            oBuf[oIdx] = 0xF0 & (temp << 4);
            oIdx++;
            incomplete = 0;
         }
         else {
            oBuf[oIdx] = newb;
            oIdx++;
         }

         newline = 1;
         mCompLines++;
      }
      else if (count < 4) {
         b1 = (oldb & 0x03);
         b2 = (count & 0x03);
         newb = (b2 << 2) | (b1);
         newb = (newb << 4) & 0xF0;
         if (incomplete == 1) {
            temp = (newb >> 4) & 0x0F;
            oBuf[oIdx] = oBuf[oIdx] | temp;
            oIdx++;
            incomplete = 0;
         }
         else {
            oBuf[oIdx] = newb;
            incomplete = 1;
         }
      }
      else if (count >= 4) {
         if (count % 4 != 0) {
            b1 = (oldb & 0x03);
            newb = 0;
            newb = ((count / 4) << 2) | b1;
            if (incomplete == 1) {
               temp = (newb >> 4) & 0x0F;
               oBuf[oIdx] = oBuf[oIdx] | temp;
               oIdx++;
               temp = newb & 0x0F;
               oBuf[oIdx] = temp << 4;
               incomplete = 1;
            }
            else {
               oBuf[oIdx] = newb;
               oIdx++;
            }

            newb = 0;
            b2 = count % 4;
            newb = (b2 << 2) | b1;
            if (incomplete == 1) {
               temp = (newb) & 0x0F;
               oBuf[oIdx] = oBuf[oIdx] | temp;
               oIdx++;
               incomplete = 0;
            }
            else {
               oBuf[oIdx] = (newb << 4) & 0xF0;
               incomplete = 1;
            }
         }
         else {
            b1 = (oldb & 0x03);
            newb = 0;
            newb = ((count / 4) << 2) | b1;
            if (incomplete == 1) {
               temp = (newb >> 4) & 0x0F;
               oBuf[oIdx] = oBuf[oIdx] | temp;
               oIdx++;
               temp = newb & 0x0F;
               oBuf[oIdx] = temp << 4;
               incomplete = 1;
            }
            else {
               oBuf[oIdx] = newb;
               oIdx++;
            }
         }
      }
      else {
         Dprintf("Error count");
      }
   }
   Dprintf("%s(%d - %d)\n", __PRETTY_FUNCTION__, oIdx, iBufSize);

   return oIdx;
}


int cOsdEm84xx::Compress4( int oBufSize )
{
   int iBufSize = (mCurwidth * mCurheight) + EM84OSD_OFFSET_BMP;
   unsigned char b,b1,b2, oldb, newb, incomplete;
   int iIdx, oIdx, count, newline, size;
   unsigned char *iBuf = mOsdBuffer;
   unsigned char *oBuf = mCOsdBuffer;

   oIdx = 0;
   mCompLines = 0;
   iIdx = EM84OSD_OFFSET_BMP;
   size = 0;
   newline = 0;
   incomplete = 0;
   b = iBuf[0];
   
   while (iIdx < iBufSize && oIdx < oBufSize)
   {
      oldb = b;
      count = 0;
      while((iIdx < iBufSize) && (count < 64) && (b == oldb) && (size < (int) mCurwidth))
      {
         oldb = b;
         b = iBuf[iIdx];
         iIdx++;
         if(newline)
         {
            oldb = b;
            newline = 0;
         }
         count++;
         size++;
      }
      if(size == (int) mCurwidth)
      {
         b1 = oldb;
         newb = 0;
         newb = newb | b1;
         size = 0;
         oBuf[oIdx] = newb;
         oIdx++;
         newline = 1;
         mCompLines++;
      }
      else if(count < 8)
      {
         newb = 0;
         b1 = (oldb);
         b2 = (count);
         newb = newb | ((b2 << 4) | (b1));
         oBuf[oIdx] = newb;
         oIdx++;
      }
      else if(count >= 8)
      {
         if(count % 8 != 0)
         {
            b1 = (oldb);
            newb = 0x80;
            newb = newb | ((((count / 8) - 1) << 4) | b1);
            oBuf[oIdx] = newb;
            oIdx++;
            newb = 0;
            b2 = count % 8;
            newb = newb | ((b2 << 4) | b1);
            oBuf[oIdx] = newb;
            oIdx++;
         }
         else
         {
            b1 = (oldb);
            newb = 0x80;
            newb = newb | ((((count / 8) - 1) << 4) | b1);
            oBuf[oIdx] = newb;
            oIdx++;
         }
      }
   }
   
   Dprintf("%s(%d - %d)\n", __PRETTY_FUNCTION__, oIdx, iBufSize);

   return oIdx;
}


void cOsdEm84xx::rgbtoyuv(double R,double G,double B,double *y,double *u,double *v)
{
   *y=(0.257*R + 0.504*G + 0.098*B) + 16.0;
   *u=(-0.148*R - 0.291*G + 0.439*B) + 128.0;
   *v=(0.439*R - 0.368*G - 0.071*B) + 128.0;
}


bool cOsdEm84xx::CheckHiliteArea( int *newColors, int numColors, cBitmap *Bitmap )
{
   int i,j,k;
   bool match = false;

   if(numColors > 4)
      return false;

   for( i = 0; i < 2; i++) {
      if( mHilitePal[0][i] == -1 ) {
         for( j = 0; j < numColors; j++ ) {
            mHilitePal[i][j] = newColors[j];
         }
         mHiliteNum[i]++;
         mHiliteArea[i].x = Bitmap->X0();
         mHiliteArea[i].y = Bitmap->Y0();
         mHiliteArea[i].w = Bitmap->Width();
         mHiliteArea[i].h = Bitmap->Height();
         return true;
      }
      for( j = 0; j < numColors; j++ ) {
         match = false;
         for( k = 0; k < 4; k++ ) {
            if( mHilitePal[i][k] == -1 || mHilitePal[i][k] == newColors[j] ) {
               match = true;
               break;
            }
         }
         if( !match ) {
            break;
         }
      }
      if( match ) {
         for( j = 0; j < numColors; j++ ) {
            for( k = 0; k < 4; k++ ) {
               if( mHilitePal[i][k] == -1 )
                  mHilitePal[i][k] = newColors[j];
            }
         }
         mHiliteNum[i]++;
         if( mHiliteNum[0] > 1 && mHiliteNum[1] > 1)
            match = false;
         else {
            // set other area to hiliteArea
            if( i == 1) {
               mHiliteIdx = 0;
               mNoHiliteIdx = 1;
            }
         }   
         break;
      }
   }
   return match;
}


cOsdProviderEm84xx::cOsdProviderEm84xx(void)
{
   Dprintf("%s\n", __PRETTY_FUNCTION__);
}


cOsd *cOsdProviderEm84xx::CreateOsd(int Left, int Top)
{
   Dprintf("%s(%d,%d)\n", __PRETTY_FUNCTION__, Left, Top);
   return new cOsdEm84xx(Left, Top);
}
